package envoy.client;

import java.time.Instant;
import java.util.Arrays;

import javax.ws.rs.client.ClientBuilder;
import javax.ws.rs.client.Entity;
import javax.ws.rs.client.WebTarget;
import javax.ws.rs.core.Response;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;

import envoy.schema.Message;
import envoy.schema.Message.MetaData.MessageState;
import envoy.schema.Messages;
import envoy.schema.ObjectFactory;
import envoy.schema.User;
import envoy.schema.Users;

/**
 * Project: <strong>envoy-client</strong><br>
 * File: <strong>Client.java</strong><br>
 * Created: <strong>28 Sep 2019</strong><br>
 * 
 * @author Kai S. K. Engelbart
 * @author Maximilian K&auml;fer
 * @author Leon Hofmeister
 * @since Envoy 0.1
 */

public class Client {

	private ObjectFactory	objectFactory	= new ObjectFactory();
	private DatatypeFactory	datatypeFactory;
	private Config			config;
	private User			sender, recipient;

	public Client(Config config, String username) {
		this.config = config;
		try {
			datatypeFactory = DatatypeFactory.newInstance();
		} catch (DatatypeConfigurationException e) {
			e.printStackTrace();
		}
		sender = getUser(username);
	}

	/**
	 * Sends a message with text content to the server.<br>
	 * Because sending a request is a blocking operation, it is executed
	 * asynchronously.
	 * 
	 * @param sender      name of the sender
	 * @param recipient   name of the recipient
	 * @param textContent content (text) of the message
	 */
	public void sendMessage(Message message) {
		new Thread(() -> {
			// Wrap single message into messages list
			Messages messages = wrapMessage(message);

			// Print message XML to console
			JAXBContext jc;
			try {
				jc = JAXBContext.newInstance("envoy.schema");
				Marshaller m = jc.createMarshaller();
				m.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, true);
				m.marshal(messages, System.out);
			} catch (JAXBException e) {
				e.printStackTrace();
			}

			// Send message
			javax.ws.rs.client.Client	client	= ClientBuilder.newClient();
			WebTarget					target	= client
				.target(String.format("%s:%d/envoy-server/rest/message/send", config.getServer(), config.getPort()));

			Response response = target.request().post(Entity.entity(messages, "application/xml"));
			System.out.println("Response code: " + response.getStatus());
			response.close();
			client.close();
		}).start();
	}

	/**
	 * Creates a {@link Message} object serializable to XML.
	 * 
	 * @param textContent The content (text) of the message
	 * @return prepared {@link Message} object
	 */
	public Message createMessage(String textContent) {
		Message.MetaData metaData = objectFactory.createMessageMetaData();
		metaData.setSender(sender.getID());
		metaData.setRecipient(recipient.getID());
		metaData.setState(MessageState.WAITING);
		metaData.setDate(datatypeFactory.newXMLGregorianCalendar(Instant.now().toString()));

		Message.Content content = objectFactory.createMessageContent();
		content.setType("text");
		content.setText(textContent);

		Message message = objectFactory.createMessage();
		message.setMetaData(metaData);
		message.getContent().add(content);

		return message;
	}

	public Messages wrapMessage(Message... messages) {
		Messages wrapper = objectFactory.createMessages();
		wrapper.getMessage().addAll(Arrays.asList(messages));
		return wrapper;
	}

	public Users getUsersListXml() {
		javax.ws.rs.client.Client	client		= ClientBuilder.newClient();
		WebTarget					target		= client
			.target(String.format("%s:%d/envoy-server/rest/user", config.getServer(), config.getPort()));
		Response					response	= target.request("application/xml").get();
		Users						users		= response.readEntity(Users.class);
		System.out.println("Response code: " + response.getStatus());
		response.close();
		client.close();
		return users;
	}

	/**
	 * Returns a {@link User} with a specific name by name.
	 * 
	 * @param name - the name of the {@link User}
	 * @return a {@link User} with the specified name
	 * @since Envoy 0.1
	 */
	private User getUser(String name) {
		javax.ws.rs.client.Client	client		= ClientBuilder.newClient();
		WebTarget					target		= client.target(String
			.format("%s:%d/envoy-server/rest/user/sender?name=%s", config.getServer(), config.getPort(), name));
		Response					response	= target.request("application/xml").get();
		Users						users		= response.readEntity(Users.class);
		System.out.println("Response code: " + response.getStatus());
		response.close();
		client.close();
		return users.getUser().get(0);
	}

	public Messages getUnreadMessages(long userId) {
		javax.ws.rs.client.Client	client			= ClientBuilder.newClient();
		WebTarget					target			= client.target(String
			.format("%s:%d/envoy-server/rest/message/receive?userId=%d", config.getServer(), config.getPort(), userId));
		Response					response		= target.request("application/xml").get();
		Messages					unreadMessages	= response.readEntity(Messages.class);
		System.out.println("Response code: " + response.getStatus());
		response.close();
		client.close();
		return unreadMessages;
	}

	/**
	 * @return the sender object that represents this client
	 * @since Envoy 0.1
	 */
	public User getSender() { return sender; }

	public User getRecipient() { return recipient; }

	public void setRecipient(User recipient) { this.recipient = recipient; }

	public boolean hasRecipient() { return recipient != null; }
}